/**
   httpGetTask.cpp
*/
/*
  (c)2023 Forward Computing and Control Pty. Ltd.
  NSW Australia, www.forward.com.au
  This code is not warranted to be fit for any purpose. You may only use it at your own risk.
  This code may be freely used for both private and commercial use
  Provided this copyright is maintained.
*/

#include <Arduino.h>
#include "volatileVars.h"
#include <HTTPClient.h>
#include "HistoryFile.h" // for struct
#include "millisDelay.h"
#include "SafeString.h"

const static float IMPORT_COST_PER_kWhr = 0.182490;
const static float EXPORT_COST_PER_kWhr =  0.070000;
const static float OFFPEAK_COST_PER_kWhr =  0.097460;
const static float DAILY_CHARGE = 1.104620;

const static unsigned long rollOver = 45; // rollover in first 45min
const static unsigned long rollOverReset = 240; // reset rollover after 4AM allow for daylight saving
const static float HeaterWattage = 3500;
const static float HeaterHysteresisWatts = 500; // need -4000 before enable heater
const static float HeaterSwitchOnValue = -(HeaterWattage + HeaterHysteresisWatts); // exports are -ve in 3PhaseWatt measurement

static millisDelay storageTimer;
static unsigned long STORAGE_TIMER_MS = 60ul * 1000; // once per min.

//#define DEBUG
static Print* debugPtr = NULL;

#define MAX_RESPONSE_SIZE 265

static HistoryRecordStruct historyRecord;

void setHttpGetTaskDebugOut(Print * out) {
  debugPtr = out;
}

float calculateTotalCost(float exportWhrs, float importWhrs, float hwWhrs, float opWhrs) {
  float totalCost = (importWhrs / 1000.0) * IMPORT_COST_PER_kWhr
                    + (exportWhrs / 1000.0) * EXPORT_COST_PER_kWhr
                    + (opWhrs / 1000.0) * OFFPEAK_COST_PER_kWhr;
  return totalCost;
}

const unsigned maxBadResponseCount = 4; // allow 4 failures => 20sec

static bool parseFloatToken(SafeString &sfToken, float &result) {
  bool rtn = true;
  // CHECK FOR 0
  sfToken.trim();
  if (sfToken.toFloat(result)) {
    // OK
  } else {
#ifdef DEBUG
    debugPtr->print(F("   ,")); debugPtr->print(sfToken); debugPtr->println(F(", is not a number"));
#endif
    rtn = false;
  }
  return rtn;
}

//struct  HistoryRecordStruct {
//  double exportWhrs;
//  double importWhrs;
//  double hotwaterWhrs;
//  double offPeakWhrs;
//  bool fullyHeated;
//  double exportWhrs_yesterday;
//  double importWhrs_yesterday;
//  double hotwaterWhrs_yesterday;
//  double offPeakWhrs_yesterday;
//  bool fullyHeated_yesterday;
//  bool midnightRestart;
// };

void rotateDayHistory(HistoryRecordStruct &historyRecord) {
  historyRecord.exportWhrs_yesterday = historyRecord.exportWhrs;
  historyRecord.exportWhrs = 0.0;
  historyRecord.importWhrs_yesterday = historyRecord.importWhrs;
  historyRecord.importWhrs = 0.0;
  historyRecord.hotwaterWhrs_yesterday = historyRecord.hotwaterWhrs;
  historyRecord.hotwaterWhrs = 0.0;
  historyRecord.offPeakWhrs_yesterday = historyRecord.offPeakWhrs;
  historyRecord.offPeakWhrs = 0.0;
  historyRecord.fullyHeated_yesterday = historyRecord.fullyHeated;
  historyRecord.fullyHeated = false;
  historyRecord.midnightRestart = true;
  writeHistory(historyRecord, debugPtr);
}

void updateHistory(float watts3Ph, float hotwater, float offPeak) {
  static millisDelay fullyHeatedDelay;
  static unsigned long fullHeatedDelay_ms = 2ul * 1000 * 60; // 2min
  static unsigned long last_ms = millis();
  unsigned long ms = millis();
  unsigned long elapsed_ms = ms - last_ms;
#ifdef DEBUG
  debugPtr->print("    updateHistory:"); debugPtr->print(elapsed_ms);
  debugPtr->print(" watts3Ph:"); debugPtr->print(watts3Ph);
  debugPtr->print(" hotwater:"); debugPtr->print(hotwater);
  debugPtr->print(" offPeak:"); debugPtr->print(offPeak);
  debugPtr->println();
#endif
  last_ms = ms;
  double extra_watts3Ph = watts3Ph * (elapsed_ms / 1000.0) / 3600.0;
  double extra_hotwater = hotwater * (elapsed_ms / 1000.0) / 3600.0;
  double extra_offPeak = offPeak * (elapsed_ms / 1000.0) / 3600.0;

  bool fullyHeated = false;
  if ((watts3Ph < HeaterSwitchOnValue) && (hotwater < 100)) {
    if (!fullyHeatedDelay.isRunning()) {
      fullyHeatedDelay.start(fullHeatedDelay_ms);
    }
  } else {
    fullyHeatedDelay.stop();
  }
  if (fullyHeatedDelay.justFinished()) {
    fullyHeated = true;
  }

  if (watts3Ph < 0) {
    historyRecord.exportWhrs += extra_watts3Ph;
  } else {
    historyRecord.importWhrs += extra_watts3Ph;
  }
  historyRecord.hotwaterWhrs += extra_hotwater;
  historyRecord.offPeakWhrs += extra_offPeak;
  historyRecord.fullyHeated = historyRecord.fullyHeated || fullyHeated;

#ifdef DEBUG
  debugPtr->print("    extra_watts3Ph:"); debugPtr->print(extra_watts3Ph, 6);
  debugPtr->print(" extra_hotwater:"); debugPtr->print(extra_hotwater, 6);
  debugPtr->print(" extra_offPeak:"); debugPtr->print(extra_offPeak, 6);
  if (fullyHeated) {
    debugPtr->print(" fullyHeated");
  }
  debugPtr->println();
#endif

  if (storageTimer.justFinished()) {
      storageTimer.repeat();
      writeHistory(historyRecord, debugPtr);  // only save once per min
   }
  // update values
  watts3PhKW_v = ((float)watts3Ph) / 1000.0;
  solarHotWaterKW_v = ((float)hotwater) / 1000.0;
  totalHotWaterKWhrs_v = (((float)historyRecord.hotwaterWhrs) + ((float)historyRecord.offPeakWhrs)) / 1000.0;
  totalCost_v = calculateTotalCost(historyRecord.exportWhrs, historyRecord.importWhrs, historyRecord.hotwaterWhrs, historyRecord.offPeakWhrs);
  fullyHeated_v = historyRecord.fullyHeated;
  updated_v = true;
}

// sfTime input 00:00 to 23:59, output day mins since 00:00
// input returned unchanged on errors
void convert24hrToMins(SafeString &sfTime, unsigned long &dayMins) {
  cSF(sfHrs, 10);
  cSF(sfMins, 10);
  if (sfTime.capacity() < 7) {
    return; // to small
  }
  if (sfTime.length() != 5) {
    return;
  }
  int idx = sfTime.indexOf(':');
  if (idx != 2) {
    return;
  }
  sfTime.substring(sfHrs, 0, idx);
  int idx2 = sfTime.indexOf(':',idx+1);
  sfTime.substring(sfMins, idx+1, idx2);
  bool isPM = false;
  int hrs = 0;
  int mins = 0;
  if (sfHrs.toInt(hrs)) {
    if (sfMins.toInt(mins)) {
      dayMins = mins + hrs * 60;
    }
  }
  return; // input unchanged
}

// returns false on error
bool parseResponse(SafeString &sfRespInit, float &watts_1, float &hotwaterWatts, float &OffPeakHW_W, unsigned long &dayMins) {

#ifdef DEBUG
  debugPtr->print(" response: "); debugPtr->println(sfRespInit);
#endif
  bool rtn = true;
  cSF(sfResp, sfRespInit.length());
  cSF(sfToken, sfRespInit.length());
  cSF(sfTime, 20);
  sfResp = sfRespInit.c_str();
  bool haveToken = sfResp.firstToken(sfToken, ','); // watts
  if (haveToken) {
    if (!parseFloatToken(sfToken, watts_1)) {
      rtn = false;
    }
  } else {
    //#ifdef DEBUG
    //    debugPtr->print("Bad response: "); debugPtr->println(sfRespInit);
    //#endif
    rtn = false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // hotwater
  if (haveToken) {
    if (!parseFloatToken(sfToken, hotwaterWatts)) {
      rtn = false;
    }
  } else {
    //#ifdef DEBUG
    //    debugPtr->print("Bad response: "); debugPtr->println(sfRespInit);
    //#endif
    rtn = false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // offPeak_hotwater
  if (haveToken) {
    if (!parseFloatToken(sfToken, OffPeakHW_W)) {
      rtn = false;
    }
  } else {
    //#ifdef DEBUG
    //    debugPtr->print("Bad response: "); debugPtr->println(sfRespInit);
    //#endif
    rtn = false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // time
  if (haveToken) {
    sfToken.trim();
    // look for contains T00: if so reset kWhrs, i.e. reset between midnight and 1am each day
    //    if (sfToken.indexOf("T00:") > 0) { // reset kWhrs  this time seems to be OK even if zero returned for the watts etc
    //      reboot();
    //    }
    int timeIdx = sfToken.indexOf("T");
    if (timeIdx > 0) {
      sfToken.substring(sfTime, timeIdx + 1, timeIdx + 6);
      convert24hrToMins(sfTime, dayMins);
    }
  } else {
    //#ifdef DEBUG
    //    debugPtr->print("Bad response: "); debugPtr->println(sfRespInit);
    //#endif
    rtn = false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // unknown
  if (haveToken) {
#ifdef DEBUG
    debugPtr->print("Bad response (too many fields): "); debugPtr->println(sfRespInit);
#endif
    rtn = false;
  }

  // else
  return rtn;
}


static void getHttpTask(void* parameter) { // parameter not used
  static unsigned badResponseCount = 0;
  float watts3Ph;
  float hotwater;
  float offPeak;

  if (!initFileSystem()) {
    if (debugPtr) {
      debugPtr->println("Start File System failed");
    }
  }
  if (!readHistory(historyRecord, &Serial)) {
    if (debugPtr) {
      debugPtr->println("Read of HistoryRecord failed");
    }
  }
  //printHistoryRecord(historyRecord, &Serial);
  storageTimer.start(STORAGE_TIMER_MS); 

  totalHotWaterKWhrs_yesterday_v = (((float)historyRecord.hotwaterWhrs_yesterday) + ((float)historyRecord.offPeakWhrs_yesterday)) / 1000.0;
  totalCost_yesterday_v = calculateTotalCost(historyRecord.exportWhrs_yesterday, historyRecord.importWhrs_yesterday, historyRecord.hotwaterWhrs_yesterday, historyRecord.offPeakWhrs_yesterday);
  fullyHeated_yesterday_v = historyRecord.fullyHeated_yesterday;
  midnightRestart_v = historyRecord.midnightRestart; // 1 if have restarted at midnight reset after 2hrs

  while (true) {
    if (wifiConnected_v) {
      HTTPClient http;

      //      if (debugPtr) {
      //        debugPtr->print("[HTTP] begin...\n");
      //      }
      http.begin("http://10.1.1.119/query?select=[Watts_3Ph.d2,Solar_HW.d2,OffPeak_HW.d2,time.local]&begin=s-5s&end=s&group=all&format=csv&missing=skip");

      //      if (debugPtr) {
      //        debugPtr->print("[HTTP] GET...\n");
      //      }
      // start connection and send HTTP header
      int httpCode = http.GET();

      // httpCode will be negative on error
      if (httpCode > 0) {
        // HTTP header has been send and Server response header has been handled
        //        if (debugPtr) {
        //          debugPtr->printf("[HTTP] GET... code: %d\n", httpCode);
        //        }

        // file found at server
        if (httpCode == HTTP_CODE_OK) {
          cSF(sfResponse, MAX_RESPONSE_SIZE);
          {
            String payload = http.getString();
            if (debugPtr) {
              debugPtr->println(payload);
            }
            sfResponse.readFrom(payload.c_str());
          }
          unsigned long dayMins = dayMins_v;
          bool responseOK = parseResponse(sfResponse, watts3Ph, hotwater, offPeak, dayMins); // parseResponse does not change sfResponse

          if (responseOK) {  // reset count
            badResponseCount = 0;
            dayMins_v = dayMins;
            if ((dayMins > rollOverReset) && (historyRecord.midnightRestart))  {
              historyRecord.midnightRestart = false;
            }
            updateHistory(watts3Ph, hotwater, offPeak);
            if ((dayMins < rollOver) && (!historyRecord.midnightRestart)) {
              rotateDayHistory(historyRecord); // and saves
              restart_v = true;
            }            
          } else {
            if (debugPtr) {
              debugPtr->print("RequestWatts Bad Data: "); debugPtr->println(sfResponse);
            }
            badResponseCount++;
          }
          if (debugPtr) {
            debugPtr->print(" Response -> watts3Ph:"); debugPtr->print(watts3Ph); debugPtr->print(" hotwater:"); debugPtr->print(hotwater);
            debugPtr->print(" offPeak:"); debugPtr->print(offPeak);
            debugPtr->print(" badResponseCount:"); debugPtr->print(badResponseCount);   debugPtr->println();
          }
        }
      } else {
        if (debugPtr) {
          debugPtr->printf("[HTTP] GET... failed, error: %s\n", http.errorToString(httpCode).c_str());
        }
      }
      http.end();
    }
    delay(5000);

  } // while(1)
}

bool createGetHttpTask() {
  BaseType_t err = xTaskCreatePinnedToCore(
                     getHttpTask,          // Task function.
                     "getHttpTask",        // name of task.
                     4 * 1024,           // Stack size of task
                     NULL,                     // parameter of the task
                     2,                        // priority of the task  // loop() has priority == 1, 2 is higher but has delay in it
                     NULL,                    // Task handle to keep track of created task
                     0);
  if (((int)err) != 1) { // pPass ,  pdTRUE      ( ( BaseType_t ) 1 ) in #include "freertos/projdefs.h"
    if (debugPtr) {
      debugPtr->print("xTaskCreatePinnedToCore returned:");
      debugPtr->println((int)err);
    }
    return false;
  }
  // could not create
  return true;
}
